import pathlib
from typing import Any

from flask import current_app

from kerko.composer import Composer
from kerko.config_helpers import config_get


def composer() -> Composer:
    return current_app.config["kerko_composer"]


def data_path() -> str:
    """
    Get the absolute path of the data directory.

    If the configuration parameter was set as a relative path, it will be
    resolved as an absolute path under the application's instance folder.
    """
    instance_path = pathlib.Path(current_app.instance_path)
    config_data_path = pathlib.Path(current_app.config.get("DATA_PATH", "kerko"))
    return str(instance_path / config_data_path)


def config(path: str) -> Any:
    """
    Retrieve an arbitrarily nested setting from the current_app's configuration.
    """
    return config_get(current_app.config, path)



import unittest

from flask import Flask

import kerko
from kerko import extractors


class LanguageExtractorTestCase(unittest.TestCase):
    """Test the `LanguageExtractor`."""

    def setUp(self):
        self.app = Flask(__name__)
        self.app.register_blueprint(kerko.make_blueprint(), url_prefix="/bibliography")
        ctx = self.app.app_context()
        ctx.push()

    @staticmethod
    def do_extract_test(language, **kwargs):
        extractor = extractors.LanguageExtractor(**kwargs)
        return extractor.extract(
            # Make fake item with the needed field.
            item={"data": {"language": language}},
            # Extractor doesn't use this arg.
            library_context={},
            # Extractor doesn't use this arg.
            spec=None,
        )

    def test_empty(self):
        result = self.do_extract_test("")
        self.assertIsNone(result)

    def test_space(self):
        result = self.do_extract_test("  ")
        self.assertIsNone(result)

    def test_separators(self):
        result = self.do_extract_test(" ; ; ")
        self.assertIsNone(result)

    def test_iso639_eng_alpha2(self):
        result = self.do_extract_test("en")
        self.assertListEqual(result, [("eng", "English")])

    def test_iso639_eng_alpha2_area(self):
        result = self.do_extract_test("en-US")
        self.assertListEqual(result, [("eng", "English")])
        result = self.do_extract_test("en_US")
        self.assertListEqual(result, [("eng", "English")])

    def test_iso639_eng_alpha3(self):
        result = self.do_extract_test("eng")
        self.assertListEqual(result, [("eng", "English")])

    def test_iso639_eng_alpha3_case_insensitive(self):
        result = self.do_extract_test("Eng")
        self.assertListEqual(result, [("eng", "English")])

    def test_iso639_eng_alpha3_trailing_spaces(self):
        result = self.do_extract_test("  eng  ")
        self.assertListEqual(result, [("eng", "English")])

    def test_iso639_eng_translated(self):
        result = self.do_extract_test("eng", locale="fr")
        self.assertListEqual(result, [("eng", "Anglais")])

    def test_iso639_eng_translated_invalid_locale(self):
        result = self.do_extract_test("eng", locale="und")
        self.assertListEqual(result, [("eng", "English")])

    def test_iso639_eng_translated_area(self):
        result = self.do_extract_test("eng", locale="fr-FR")
        self.assertListEqual(result, [("eng", "Anglais")])

    def test_iso639_eng_normalize_invalid(self):
        result = self.do_extract_test("eng", normalize_invalid=str.upper)
        self.assertListEqual(result, [("eng", "English")])

    def test_iso639_eng_name(self):
        result = self.do_extract_test("English")
        self.assertListEqual(result, [("eng", "English")])

    def test_iso639_eng_name_case_insensitive(self):
        result = self.do_extract_test("ENGLISH")
        self.assertListEqual(result, [("eng", "English")])

    def test_iso639_fra_alpha3(self):
        result = self.do_extract_test("fra")
        self.assertListEqual(result, [("fra", "French")])

    def test_iso639_fra_alpha3_bibliographic(self):
        result = self.do_extract_test("fre")
        self.assertListEqual(result, [("fra", "French")])

    def test_iso639_fra_name(self):
        result = self.do_extract_test("French")
        self.assertListEqual(result, [("fra", "French")])

    def test_unknown_name(self):
        result = self.do_extract_test("Newspeak")
        self.assertListEqual(result, [("newspeak", "Newspeak")])

    def test_unknown_normalize_invalid(self):
        result = self.do_extract_test("Newspeak", normalize_invalid=str.upper)
        self.assertListEqual(result, [("newspeak", "NEWSPEAK")])

    def test_unknown_name_area(self):
        result = self.do_extract_test("xx-XX")
        self.assertListEqual(result, [("xx-xx", "Xx-Xx")])
        result = self.do_extract_test("xx_XX")
        self.assertListEqual(result, [("xx_xx", "Xx_Xx")])

    def test_unknown_name_invalid(self):
        result = self.do_extract_test("Newspeak", allow_invalid=False)
        self.assertEqual(result, None)

    def test_multiple(self):
        result = self.do_extract_test("eng ; fre; por")
        self.assertListEqual(result, [("eng", "English"), ("fra", "French"), ("por", "Portuguese")])

    def test_multiple_separator_re(self):
        result = self.do_extract_test("eng / fre:por |ita ben", values_separator_re=r"[, |/:]")
        self.assertListEqual(
            result,
            [
                ("eng", "English"),
                ("fra", "French"),
                ("por", "Portuguese"),
                ("ita", "Italian"),
                ("ben", "Bengali"),
            ],
        )

    def test_multiple_valid_and_invalid(self):
        result = self.do_extract_test("th; unknown1; unknown2; hat", allow_invalid=False)
        self.assertListEqual(result, [("tha", "Thai"), ("hat", "Haitian")])

    def test_multiple_translated(self):
        result = self.do_extract_test("it; fr", locale="fr")
        self.assertListEqual(result, [("ita", "Italien"), ("fra", "Français")])

    def test_multiple_duplicated(self):
        result = self.do_extract_test("eng; eng; en; English; en-GB")
        self.assertListEqual(result, [("eng", "English")])

    def test_no_normalize(self):
        result = self.do_extract_test("en", normalize=False)
        self.assertListEqual(result, [("en", "en")])

    def test_no_normalize_multiple(self):
        result = self.do_extract_test("eng; fra", normalize=False)
        self.assertListEqual(result, [("eng", "eng"), ("fra", "fra")])

    def test_no_normalize_duplicates(self):
        result = self.do_extract_test("en; en", normalize=False)
        self.assertListEqual(result, [("en", "en")])


class CollectionNamesExtractorTestCase(unittest.TestCase):
    """Test the `CollectionNamesExtractor`."""

    class FakeLibraryContext:
        """Mock the interface of `kerko.sync.zotero.LibraryContext`."""

        def __init__(self, collections):
            self.collections = collections

    def setUp(self):
        self.app = Flask(__name__)
        self.app.register_blueprint(kerko.make_blueprint(), url_prefix="/bibliography")
        ctx = self.app.app_context()
        ctx.push()

    @classmethod
    def do_extract_test(cls, library_collections, item_collections, **kwargs):
        extractor = extractors.CollectionNamesExtractor(**kwargs)
        return extractor.extract(
            # Build fake item data.
            item={"data": {"collections": item_collections}},
            # Build fake library data.
            library_context=cls.FakeLibraryContext(
                collections={
                    key: {"data": {"name": name}} for key, name in library_collections.items()
                }
            ),
            # Extractor doesn't use this arg.
            spec=None,
        )

    def test_empty(self):
        library_collections = {}
        item_collections = []
        result = self.do_extract_test(library_collections, item_collections)
        self.assertIsNone(result)

    def test_none(self):
        library_collections = {"0": "a", "1": "b"}
        item_collections = []
        result = self.do_extract_test(library_collections, item_collections)
        self.assertIsNone(result)

    def test_unknown(self):
        library_collections = {}
        item_collections = ["0"]
        result = self.do_extract_test(library_collections, item_collections)
        self.assertIsNone(result)

    def test_single(self):
        library_collections = {"0": "a", "1": "b", "2": "c"}
        item_collections = ["0"]
        result = self.do_extract_test(library_collections, item_collections)
        self.assertEqual(result, "a")

    def test_multiple_empty(self):
        library_collections = {"0": "", "1": " ", "2": "   "}
        item_collections = ["0", "1", "2"]
        result = self.do_extract_test(library_collections, item_collections)
        self.assertIsNone(result)

    def test_multiple_sorted(self):
        library_collections = {
            "x0": "ggg",
            "40": "a",
            "01": "Y",
            "22": "zz",
            "32": "x",
            "x1": "hhh",
            "24": "f",
            "15": "bbb",
            "06": "CCC",
            "17": "d",
            "08": "E",
        }
        item_collections = ["40", "01", "22", "32", "24", "15", "06", "17", "08"]
        result = self.do_extract_test(library_collections, item_collections)
        self.assertEqual(
            result,
            extractors.RECORD_SEPARATOR.join(["a", "bbb", "CCC", "d", "E", "f", "x", "Y", "zz"]),
        )

    def test_strip(self):
        library_collections = {
            "0": " a",
            "1": "b ",
            "2": " c ",
            "3": "\nd\n",
            "4": " e e ",
            "5": " ",
        }
        item_collections = ["0", "1", "2", "3", "4", "5"]
        result = self.do_extract_test(library_collections, item_collections)
        self.assertEqual(
            result,
            extractors.RECORD_SEPARATOR.join(["a", "b", "c", "d", "e e"]),
        )


class TagsTextExtractorTestCase(unittest.TestCase):
    """Test the `TagsTextExtractor`."""

    def setUp(self):
        self.app = Flask(__name__)
        self.app.register_blueprint(kerko.make_blueprint(), url_prefix="/bibliography")
        ctx = self.app.app_context()
        ctx.push()

    @staticmethod
    def do_extract_test(tags, **kwargs):
        extractor = extractors.TagsTextExtractor(**kwargs)
        return extractor.extract(
            # Build fake item data.
            item={"data": {"tags": [{"tag": tag} for tag in tags]}},
            # Extractor doesn't use this arg.
            library_context={},
            # Extractor doesn't use this arg.
            spec=None,
        )

    def test_empty(self):
        tags = []
        result = self.do_extract_test(tags)
        self.assertIsNone(result)

    def test_single(self):
        tags = ["a"]
        result = self.do_extract_test(tags)
        self.assertEqual(result, "a")

    def test_multiple_empty(self):
        tags = ["", " ", "   "]
        result = self.do_extract_test(tags)
        self.assertIsNone(result)

    def test_multiple_sorted(self):
        tags = ["a", "Y", "zz", "x", "f", "bbb", "CCC", "d", "E"]
        result = self.do_extract_test(tags)
        self.assertEqual(
            result,
            extractors.RECORD_SEPARATOR.join(["a", "bbb", "CCC", "d", "E", "f", "x", "Y", "zz"]),
        )

    def test_strip(self):
        tags = [" a", "b ", " c ", " ", "\nd\n", " e e "]
        result = self.do_extract_test(tags)
        self.assertEqual(
            result,
            extractors.RECORD_SEPARATOR.join(["a", "b", "c", "d", "e e"]),
        )

    def test_include_re(self):
        tags = ["a", "_b", "_c", "d", "ex", "x", "xx"]
        result = self.do_extract_test(tags, include_re=r"((^_)|(.*x))")
        self.assertEqual(
            result,
            extractors.RECORD_SEPARATOR.join(["_b", "_c", "ex", "x", "xx"]),
        )

    def test_exclude_re(self):
        tags = ["a", "_b", "_c", "d", "ex", "x", "xx"]
        result = self.do_extract_test(tags, exclude_re=r"((^_)|(.*x))")
        self.assertEqual(
            result,
            extractors.RECORD_SEPARATOR.join(["a", "d"]),
        )

    def test_include_exclude_re(self):
        tags = ["a", "_b", "_c", "d", "ex", "x", "xx"]
        result = self.do_extract_test(tags, include_re=r"(^_)", exclude_re=r"(.*x)")
        self.assertEqual(
            result,
            extractors.RECORD_SEPARATOR.join(["_b", "_c"]),
        )



"""
Tests for the Atom feed view.
"""

import re
from datetime import datetime

from flask import url_for
from lxml import etree
from werkzeug.datastructures import MultiDict

from kerko.criteria import create_feed_criteria
from kerko.searcher import Searcher
from kerko.shortcuts import config
from kerko.storage import open_index
from tests.integration_testing import SynchronizedTestCase


class AtomFeedTestCase(SynchronizedTestCase):
    """Test the Atom feed."""

    TEST_ITEMS_COUNT = 6  # Update when number of records in test dataset changes.

    def setUp(self):
        self.namespaces = {
            "ns": "http://www.w3.org/2005/Atom",
            "dc": "http://purl.org/dc/elements/1.1/",
        }

    def test_main_feed(self):
        """Test the main feed."""

        # TODO: Test multipage feed (need more entries in test dataset).

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/atom.xml")
            self.assertEqual(response.status_code, 200)

            # Check feed elements.
            root = etree.fromstring(response.get_data(as_text=True).encode("utf-8"))
            self.assertEqual(
                root.xpath("/ns:feed/ns:id/text()", namespaces=self.namespaces)[0],
                url_for(
                    ".atom_feed",
                    _external=True,
                    **create_feed_criteria().params(options={"page": None}),
                ),
                "feed.id is invalid",
            )
            self.assertRegex(
                root.xpath("/ns:feed/ns:title/text()", namespaces=self.namespaces)[0],
                rf"^{re.escape(config('kerko.meta.title'))}",
                "feed.title is incorrect",
            )
            # Check date, with a compromise: not trying full ISO 8601 variants.
            self.assertIsInstance(
                datetime.strptime(
                    root.xpath("/ns:feed/ns:updated/text()", namespaces=self.namespaces)[0][0:13],
                    "%Y-%m-%dT%H",
                ),
                datetime,
                "feed.updated is invalid",
            )

            # Check feed entry elements.
            entry_count = len(root.xpath("/ns:feed/ns:entry", namespaces=self.namespaces))
            self.assertEqual(
                entry_count,
                self.TEST_ITEMS_COUNT,
            )

            # Check elements under each entry element. A compromise to reduce
            # tests maintenance: not matching with actual items, just checking
            # the presence and format of the values.
            for position in range(1, self.TEST_ITEMS_COUNT + 1):
                entry_id = root.xpath(
                    f"/ns:feed/ns:entry[position() = {position}]/ns:id/text()",
                    namespaces=self.namespaces,
                )[0]
                self.assertRegex(
                    entry_id,
                    r"^http://.+/[A-Z0-9]+$",
                    f"feed.entry[{position}].id is invalid",
                )
                entry_href = root.xpath(
                    f"/ns:feed/ns:entry[position() = {position}]/ns:link/@href",
                    namespaces=self.namespaces,
                )[0]
                self.assertRegex(
                    entry_href,
                    r"^http://.+/[A-Z0-9]+$",
                    f"feed.entry[{position}].link is invalid",
                )
                self.assertEqual(entry_id, entry_href)
                self.assertFalse(
                    Searcher(open_index("index"))
                    .search(
                        require_all={"id": re.search(r"/([A-Z0-9]+)$", entry_id).groups(1)},
                        limit=1,
                    )
                    .is_empty(),
                )
                entry_title = root.xpath(
                    f"/ns:feed/ns:entry[position() = {position}]/ns:title/text()",
                    namespaces=self.namespaces,
                )[0]
                self.assertGreater(
                    len(entry_title),
                    0,
                    f"feed.entry[{position}].title is empty",
                )
                entry_dc_title = root.xpath(
                    f"/ns:feed/ns:entry[position() = {position}]/dc:title/text()",
                    namespaces=self.namespaces,
                )[0]
                self.assertGreater(
                    len(entry_dc_title),
                    0,
                    f"feed.entry[{position}].dc:title is empty",
                )
                self.assertEqual(entry_title, entry_dc_title)
                # Check dates, with a compromise: not trying full ISO 8601 variants.
                self.assertIsInstance(
                    datetime.strptime(
                        root.xpath(
                            f"/ns:feed/ns:entry[position() = {position}]/ns:published/text()",
                            namespaces=self.namespaces,
                        )[0][0:13],
                        "%Y-%m-%dT%H",
                    ),
                    datetime,
                    f"feed.entry[{position}].published is invalid",
                )
                self.assertIsInstance(
                    datetime.strptime(
                        root.xpath(
                            f"/ns:feed/ns:entry[position() = {position}]/ns:updated/text()",
                            namespaces=self.namespaces,
                        )[0][0:13],
                        "%Y-%m-%dT%H",
                    ),
                    datetime,
                    f"feed.entry[{position}].updated is invalid",
                )

    def test_search_feed(self):
        """Test a search feed."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/atom.xml?all=4507457-YE4WVE35")
            self.assertEqual(response.status_code, 200)

            root = etree.fromstring(response.get_data(as_text=True).encode("utf-8"))
            self.assertEqual(len(root.xpath("/ns:feed/ns:entry", namespaces=self.namespaces)), 1)
            self.assertEqual(
                root.xpath("/ns:feed/ns:entry/ns:title/text()", namespaces=self.namespaces)[0],
                "How to Make a Faceted Classification and Put It On the Web",
            )
            self.assertEqual(
                root.xpath("/ns:feed/ns:entry/dc:title/text()", namespaces=self.namespaces)[0],
                "How to Make a Faceted Classification and Put It On the Web",
            )
            self.assertEqual(
                root.xpath("/ns:feed/ns:entry/dc:creator/text()", namespaces=self.namespaces)[0],
                "Denton, William",
            )
            self.assertEqual(
                root.xpath("/ns:feed/ns:entry/dc:date/text()", namespaces=self.namespaces)[0],
                "2003",
            )
            self.assertEqual(
                root.xpath("/ns:feed/ns:entry/ns:published/text()", namespaces=self.namespaces)[0],
                "2021-11-13T23:10:11Z",
            )

    def test_empty_search_feed(self):
        """Test an empty search feed."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/atom.xml?all=qwertyasdfghzxcvbn")
            self.assertEqual(response.status_code, 200)

            root = etree.fromstring(response.get_data(as_text=True).encode("utf-8"))
            self.assertEqual(
                root.xpath("/ns:feed/ns:id/text()", namespaces=self.namespaces)[0],
                url_for(
                    ".atom_feed",
                    _external=True,
                    **create_feed_criteria(
                        initial=MultiDict({"all": ["qwertyasdfghzxcvbn"]})
                    ).params(options={"page": None}),
                ),
                "feed.id is missing or incorrect",
            )
            self.assertRegex(
                root.xpath("/ns:feed/ns:title/text()", namespaces=self.namespaces)[0],
                rf"^{re.escape(config('kerko.meta.title'))}",
                "feed.title is missing or incorrect",
            )
            self.assertIsInstance(
                datetime.strptime(
                    root.xpath("/ns:feed/ns:updated/text()", namespaces=self.namespaces)[0][0:13],
                    "%Y-%m-%dT%H",  # A compromise: not trying all ISO 8601 variants.
                ),
                datetime,
                "feed.updated is missing or incorrectly formatted",
            )
            self.assertEqual(
                len(root.xpath("/ns:feed/ns:entry", namespaces=self.namespaces)),
                0,
                "Feed should have no entry element",
            )



"""Text utilities."""

# Instead of adding more dependencies, let's take advantage of Whoosh's text
# processing utilities.
from whoosh.analysis.filters import CharsetFilter, LowercaseFilter
from whoosh.analysis.tokenizers import IDTokenizer, RegexTokenizer
from whoosh.support.charset import accent_map

ID_NORMALIZATION_CHAIN = IDTokenizer() | CharsetFilter(accent_map) | LowercaseFilter()
SORT_NORMALIZATION_CHAIN = RegexTokenizer() | CharsetFilter(accent_map) | LowercaseFilter()
SLUGIFY_CHAIN = RegexTokenizer(r"[^_\W]+") | CharsetFilter(accent_map) | LowercaseFilter()


def id_normalize(text):
    return " ".join([t.text for t in ID_NORMALIZATION_CHAIN(text)])


def sort_normalize(text):
    return " ".join([t.text for t in SORT_NORMALIZATION_CHAIN(text)])


def slugify(text, delimiter="-"):
    return delimiter.join([t.text for t in SLUGIFY_CHAIN(text)])



"""Tags-related utilities."""

import re
from collections.abc import Iterable


class TagGate:
    """Determine whether an object is included or excluded based on its tags."""

    def __init__(self, include_re="", exclude_re=""):
        """
        Initialize the instance.

        :param [str,list] include_re: Regular expression pattern to use to
            include objects based on their tags. Any object which does not have
            a tag that matches this pattern will be excluded. If empty (which is
            the default), all objects will be included unless the `exclude_re`
            argument is set and causes some to be excluded. When passing a list,
            every pattern of the list must match at least a tag for the object
            to be included.

        :param [str,list] exclude_re: Regular expression pattern to use to
            exclude objects based on their tags. Any object that have a tag that
            matches this pattern will be excluded. If empty (which is the
            default), no objects will be excluded unless the `include_re`
            argument is set, in which case items that don't have any tag that
            matches it will be excluded. When passing a list, every pattern of
            the list must match at least a tag for the object to be excluded.
        """
        if include_re:
            assert isinstance(include_re, Iterable)
            if isinstance(include_re, str):
                include_re = [include_re]
            self.include_re = [re.compile(pattern) for pattern in include_re]
        else:
            self.include_re = None

        if exclude_re:
            assert isinstance(exclude_re, Iterable)
            if isinstance(exclude_re, str):
                exclude_re = [exclude_re]
            self.exclude_re = [re.compile(pattern) for pattern in exclude_re]
        else:
            self.exclude_re = None

    def check(self, obj):
        """
        Check whether the object is to be included or excluded.

        :param dict obj: A dict which is expected to contain an optional 'tags'
            key, whose corresponding value should be a list of dicts, each with
            a 'tag' key, whose corresponding value should be a string
            representing a tag.
        """
        included = [True] if not self.include_re else [False] * len(self.include_re)
        excluded = [False] if not self.exclude_re else [False] * len(self.exclude_re)
        if self.include_re:
            self._check_expressions(obj, self.include_re, included)
        if self.exclude_re:
            self._check_expressions(obj, self.exclude_re, excluded)
        return all(included) and not all(excluded)

    @staticmethod
    def _check_expressions(obj, expressions, expr_results):
        for i, expr in enumerate(expressions):
            for tag_data in obj.get("tags", []):
                tag = tag_data.get("tag", "").strip()
                if expr.match(tag):
                    expr_results[i] = True
                    break
